%%Lab 6: Probability 
clear all; close all;

%% Let's start with summary statistics- we looked at a few: mean, median, mode that were important. 
% What is the intuition behind these? (Explain on board) - statistic is a
% single number and we want that single number to represent the data well
% can measure this by looking at discrepancies from all data points using a
% metric. What metric? ---> norms  --> l0, l1, l2 ... lp

%let's say we have a single data point x_i = 1...
% Plot the l0,l1, l2 norm discrepancies to a statistic s. 
% use s = linspace(-1,3,41) 

x_i = 1;
s = linspace(-1,3,41);
ind = find(s==x_i);
l0norm = ones(1,41);
l0norm(ind) = 0;

plot(s, l0norm)
hold on;

l1norm = abs(x_i - s);
plot(s, l1norm)

l2norm = (x_i - s).^2;
plot(s,l2norm)

legend('l0', 'l1', 'l2')
xlabel('Summary Statistic Value')
ylabel('Discrepancy to data point x_i = 1')

%In this case the statistic s = x_1 - this is obvious but the shape of the
%curves gives you some intuition about what each norm statistic means. Now
%we can think about this in aggregate over a bunch of data points --> on
%board --> leads us to mean, median, and mode
% say x = [0.5,1,1,1,3]- compute the aggregate norms and plot for s =
% linspace(-1,3,41)

x = [0.5,0.5,1,1,4];
l0norm_sum = zeros(1,41);
l1norm_sum = zeros(1,41);
l2norm_sum = zeros(1,41);
for i = 1:length(x)
    x_i = x(i);
    ind = find(s==x_i);
    l0norm = ones(1,41);
    l0norm(ind) = 0;
    l0norm_sum = l0norm_sum + l0norm;
    
    l1norm = abs(x_i-s);
    l1norm_sum = l1norm_sum + l1norm;
    
    l2norm = (x_i-s).^2;
    l2norm_sum = l2norm_sum + l2norm;
end

figure; 
plot(s, l0norm_sum)
hold on;

[minv,mini] = min(l0norm_sum);
scatter(s(mini),minv,'*')

plot(s, l1norm_sum)
[minv,mini] = min(l1norm_sum);
scatter(s(mini),minv,'*')

plot(s,l2norm_sum)
[minv,mini] = min(l2norm_sum);
scatter(s(mini),minv,'*')

legend('l0','l0min', 'l1','l1min', 'l2', 'l2min')
xlabel('Summary Statistic Value')
ylabel('Discrepancy to data points')

%We see here the mode and the median provide equally good estimates but the
%mean is thrown off by the outlier 4. This gets to the point that any
%summary statistic may be good or bad depending on your dataset. We can see
%this again in a more real example comparing a gaussian to a skewed
%distribution (randn, and pearsrnd). 

clear all; close all;
x = randn(10000,1);
histogram(x)
hold on;
meanval = mean(x);
medianval = median(x);
line([meanval, meanval], ylim, 'LineWidth', 2, 'Color', 'r');
line([medianval, medianval], ylim, 'LineWidth', 2, 'Color', 'c');
legend('datahist','mean','median')

figure;
x = pearsrnd(0,1,0.8,3,10000,1);
histogram(x)
hold on;
meanval = mean(x);
medianval = median(x);
line([meanval, meanval], ylim, 'LineWidth', 2, 'Color', 'r');
line([medianval, medianval], ylim, 'LineWidth', 2, 'Color', 'c');
legend('datahist','mean','median')

% The median here is a better statistic for the skewed distribution even
% though for a normal distribution the mean and median are both equally
% good.

%% Let's think more about the summary statistics involving mean and variance
% Take a die that has a uniform probability of showing one of the 6 sides
% What is the distribution of X?  Uniform
clear all; close all;
x = 1:1:6;
y = [1/6,1/6,1/6,1/6,1/6,1/6];
plot(x,y);

%How do we generate samples? go through sampling on board..
% Since you implement sampling on the hw let's use the function
% randsample(n,k) which generates k samples from 1:n
% generate 50 samples and histogram the output 
samps = randsample(6,50,1);
histogram(samps);

%calculate sample mean is this the expectation? why? What about sample variance? E[x] =
%1/6+1/3+1/2+2/3+5/6+1 = 3.5 Var(X) = 105/36 = 2.9167 
sampmean = sum(samps)/50
sampvar = sum((samps-sampmean).^2)/50

%Do this in a loop and compute the average of the sample means and the
%average of the sample variances over 100000 different generations of samples
avgsampmean = 0;
avgsampvar = 0;
for j = 1:100000
    samps = randsample(6,50,1);
    sampmean=sum(samps/50);
    avgsampmean = avgsampmean + sampmean;
    avgsampvar = avgsampvar + sum((samps-sampmean).^2)/50;
end
avgsampmean = avgsampmean/100000
avgsampvar = avgsampvar/100000

%On board discuss sample statistics --> biased vs. unbiased estimators- so
%now go back and try the unbiased sample variance
avgsampmean = 0;
avgsampvar = 0;
for j = 1:100000
    samps = randsample(6,50,1);
    sampmean=sum(samps/50);
    avgsampmean = avgsampmean + sampmean;
    avgsampvar = avgsampvar + sum((samps-sampmean).^2)/49;
end
avgsampmean = avgsampmean/100000
avgsampvar = avgsampvar/100000

%What about N = 100000 now 
samps = randsample(6,1000000,1);
sampmean = sum(samps)/1000000
sampvar = sum((samps-sampmean).^2)/1000000

%% Bias Variance Tradeoff (On board / discussion)


%% CLT 
% Why does the CLT matter? Let's think about all the noise sources in an
% experiment.. let's say we have an electrical system that is composed of
% many circuit elements and connections. Shot noise is a common type of
% noise that is drawn from a poisson process (i.e.  subsequent time intervals of . Let's say we have a signal traveling through 
% electrical components and whenever there is a new circuit connection we add shot noise (drawn from a poisson distribution)
% generate a random
% sample of length 100 from a poisson distribution to indicate 1 noise
% source that generates noise in this system at a 100 time points such that
% over the 100 time points the noise is poisson distributed
% 
clear all; close all;

random_sample1 = poissrnd(2,1,100);
figure;
plot(random_sample1)
figure;
histogram(random_sample1)

%Now say there are many components that each are adding this shot noise 
% The total noise at the output of the system will be the sum of these
% noise sources so we can think of the output as the sum of random samples
% from poisson distributions. Generate a 100 dimensional noise sample that
% is the sum of these 100 noise sources. Normalize the sum by the total
% number of noise sources

noiseoutput = zeros(1,100);
for source = 1:100
    noiseoutput = noiseoutput + poissrnd(2,1,100);
end

figure; 
noiseoutput = noiseoutput/100;
plot(noiseoutput);
figure;
histogram(noiseoutput)

%because of the CLT the final noise looks Gaussian- this is why in a lot of
%experiments we assume the final noise source to be white- it is the sum of
%many different noise sources of different distributions but this sum ends
%up being gaussian.

%% PCA whitening

% Removing correlations and normalizing features is an important piece of
% many learning algorithms and is something that the visual system must do
% because we would like to remove redundant information and make each
% feature as informative as possible- also make data such taht the features
% all have unit variance. 

% load the data and measure it's covariance - then whiten the data
clear all; close all;
load gausscov.mat
figure
plot(X(1,:),X(2,:),'+')
covmat = cov(X');

%How do we whiten this? i.e. turn this data into data that has covariance =
%identity: Y = MX --> I = MCxMt --> I = MVDVtM --> MV(D^1/2)(D^1/2)VtMt -->
%M = (D^(-1/2)*V^T)
%

%rotate the data to decorrelate it using the PCA axes
[V,D] = eig(covmat);
rotDat = V'*X;
figure; 
plot(rotDat(1,:), rotDat(2,:), '+')
cov(rotDat')

%scale the variances of each feature to be the same
Dsqinv = [1/sqrt(D(1,1)),0 ; 0, 1/sqrt(D(2,2))];
whitenedDat = Dsqinv*rotDat;
figure; 
plot(whitenedDat(1,:), whitenedDat(2,:), '+')
cov(whitenedDat')

% This gives us an interpretation of PCA as learning axes that
% statistically decorrelate the data 







